//=======================================================================
//           This file is a part of the Juju GUI application.
//                  Copyright (C) 2011 Nathan Osman
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include <CEnvironmentsDialog.h>
#include "ui_CEnvironmentsDialog.h"

// Qt includes
#include <QDebug>
#include <QDir>
#include <QFile>
#include <QMessageBox>

// C++ includes
#include <fstream>

// Juju GUI includes
#include <CEC2EnvironmentWidget.h>
#include <CGenericEnvironmentWidget.h>
#include <CLocalEnvironmentWidget.h>

CEnvironmentsDialog::CEnvironmentsDialog(QWidget * parent)
    : QDialog(parent), ui(new Ui::CEnvironmentsDialog)
{
    ui->setupUi(this);

    ui->Environments->setBaseSize(100, 100);

    // Determine the absolute filename of the config file
    QString config_file = QDir::home().absoluteFilePath(".juju/environments.yaml");
    m_config_file = config_file.toUtf8().data();

    // Check to see if the config file exists
    if(QFile::exists(config_file))
    {
        // Attempt to load the file
        if(LoadConfig())
            PopulateListbox();
        else
            QMessageBox::critical(this,
                                  tr("Error:"),
                                  tr("An error occurred while loading %1:\n\n%2")
                                      .arg(config_file).arg(m_error));
    }
    else
        qDebug() << tr("The configuration file %1 does not exist.").arg(config_file);
}

CEnvironmentsDialog::~CEnvironmentsDialog()
{
    delete ui;
}

/// Called when the Add button is clicked to add an environment.
void CEnvironmentsDialog::OnAdd()
{
    //...
}

/// Called when the Remove button is clicked to remove an environment.
void CEnvironmentsDialog::OnRemove()
{
    //...
}

/// Called when the Set Default button is clicked to set the default environment.
void CEnvironmentsDialog::OnSetDefault()
{
    // Clear the old default
    if(!m_default_environment.isEmpty())
        ToggleDefault(m_default_environment, false);

    // The default environment is now the currently selected one
    m_default_environment = ui->Environments->currentItem()->text();

    // Set its name accordingly
    ToggleDefault(m_default_environment, true);
}

/// Called when the current selection in the environments listbox changes.
/**
  * \param current the currently selected item
  */
void CEnvironmentsDialog::OnItemChanged(QListWidgetItem * current, QListWidgetItem *)
{
    // Get the name of the environment we are switching to
    QString env_name = current->data(Qt::UserRole).toString();

    // Retrieve the type of that environment
    EnvironmentMap::iterator i = m_environments.find(env_name);
    if(i != m_environments.end())
    {
        // Delete the old widget
        delete ui->EnvironmentWidget;

        /*
        if(i.value()["type"] == "local")
            ui->Layout->addWidget(ui->EnvironmentWidget = new CLocalEnvironmentWidget(this));
        else if(i.value()["type"] == "ec2")
            ui->Layout->addWidget(ui->EnvironmentWidget = new CEC2EnvironmentWidget(this));
        else
            ui->Layout->addWidget(ui->EnvironmentWidget = new CGenericEnvironmentWidget(this));
        */

        ui->Layout->addWidget(ui->EnvironmentWidget = new CGenericEnvironmentWidget(i.value(), this));
    }
}

/// Attempts to load the configuration from the file in the user's home directory.
/**
  * \return true if the configuration was loaded without error
  */
bool CEnvironmentsDialog::LoadConfig()
{
    // Begin by attempting to open the config file
    std::ifstream fin(m_config_file);
    if(!fin.is_open())
    {
        m_error = tr("Error opening the configuration file for reading.");
        return false;
    }

    // Now begin parsing the document
    try
    {
        // Juju never uses more than one document
        YAML::Parser parser(fin);
        YAML::Node document;

        if(!parser.GetNextDocument(document))
        {
            m_error = tr("Error parsing YAML document.");
            return false;
        }

        // We only recognize two keys 'default' and 'environments'
        // - the former is optional and indicates the default environment

        // Make sure the environments are specified
        const YAML::Node * environments = document.FindValue("environments");
        if(!environments)
        {
            m_error = tr("'environments' key missing from YAML document.");
            return false;
        }

        // Enumerate the environments we find
        for(YAML::Iterator i=environments->begin();i!=environments->end();++i)
        {
            // Determine what environment we are parsing
            const YAML::Node * env_type = i.second().FindValue("type");
            if(!env_type)
            {
                m_error = tr("Environment type not specified.");
                return false;
            }

            // Get the name of the environment
            std::string env_name;
            i.first() >> env_name;

            // Load the environment's configuration
            std::string env_type_str;
            if(!env_type->GetScalar(env_type_str))
            {
                m_error = tr("Expected scalar value for 'type' entry.");
                return false;
            }

            if(env_type_str == "ec2" || env_type_str == "local")
                AddEnvironmentToQMap(env_name, i.second());
            else
                qDebug() << tr("Warning: unknown environment type '%1'.").arg(env_type_str.c_str());
        }

        // Check for the 'default' key and set it otherwise
        const YAML::Node * default_env = document.FindValue("default");
        if(default_env)
        {
            // Try to find the specified environment
            std::string default_env_str;
            default_env->GetScalar(default_env_str);

            m_default_environment = default_env_str.c_str();
        }
        // Otherwise grab the first environment
        else if(m_environments.size())
            m_default_environment = m_environments.constBegin().key();
        else
            qDebug() << tr("Warning: no environments found - no default environment set.");
    }
    catch(YAML::ParserException & exception)
    {
        // Store a copy of the error message encountered
        m_error = exception.what();
        return false;
    }

    return true;
}

/// Attempts to store the configuration in the user's home configuration.
/**
  * \return true if the configuration was stored on disk without error
  */
bool CEnvironmentsDialog::StoreConfig()
{
    //...
    return true;
}

/// Adds the specified configuration to the list of environments.
/**
  * \param env_name the name of the environment
  * \param env_conf a map containing the configuration for the environment
  */
void CEnvironmentsDialog::AddEnvironmentToQMap(std::string & env_name, const YAML::Node & env_conf)
{
    // This map will contain the key:value pairs for the environment
    StringMap env_map;
    for(YAML::Iterator i=env_conf.begin();i!=env_conf.end();++i)
    {
        std::string key, value;
        i.first() >> key;
        i.second() >> value;

        env_map.insert(key.c_str(), value.c_str());
    }

    // Add the map to the specified name
    m_environments.insert(env_name.c_str(), env_map);
}

/// Populates the listbox with all of the environments.
void CEnvironmentsDialog::PopulateListbox()
{
    for(EnvironmentMap::const_iterator i = m_environments.begin(); i != m_environments.end(); ++i)
    {
        QListWidgetItem * item = new QListWidgetItem(ui->Environments);
        item->setText(i.key());
        item->setData(Qt::UserRole ,i.key());

        ui->Environments->addItem(item);
    }

    // Set the default item
    if(m_environments.size())
        ToggleDefault(m_default_environment, true);
}

/// Toggles whether the selected list entry is the default or not.
/**
  * \param env_name the name of the environment to toggle
  * \param set_default true to set the environment as the default
  *
  * Note: this only affects the list box and not the configuration file.
  */
void CEnvironmentsDialog::ToggleDefault(QString env_name, bool set_default)
{
    QList<QListWidgetItem *> matched_items = ui->Environments->findItems(env_name + ((set_default)?"":tr(" [default]")),
                                                                         Qt::MatchExactly);
    if(matched_items.size())
        matched_items.at(0)->setText(env_name + ((set_default)?tr(" [default]"):""));
}
